/*-
 * Copyright (c) 2001-2003 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "compat.h"

#ifndef lint
static const char rcsid[] ATTR_UNUSED =
  "@(#)$Id: kip6fw.c,v 1.3 2006/04/26 20:43:43 simon Exp $";
#endif /* !lint */

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/queue.h>

#include <osreldate.h>

#include <ipa_mod.h>

#include "ipa_ip6fw.h"
#include "kip6fw.h"

#if __FreeBSD_version < 420001
/*
 * There were bugs in IPv6 Firewall implementation since FreeBSD
 * 4.0-RELEASE (see CVS log for src/sbin/ip6fw.c), starting from
 * FreeBSD 4.2-RELEASE those bugs were fixed.  Magic 1024 number
 * came from the early src/sbin/ip6fw.c:list() source.
 */

# define IP6FW_INCREASE_ENTRIES	1024	/* How many entries allocate in fw_buf. */

#else

# define IP6FW_INCREASE_ENTRIES	20	/* How many entries allocate in fw_buf. */
# define IP6FW_FREE_ENTRIES	30	/* Max allowed free entries in fw_buf. */

#endif /* __FreeBSD_version < 420001 */

/*
 * List of active kip6fw_rules.  Since all kip6fw_rules use
 * marray, then they are placed in one chunk of memory, let's
 * keep active rules list sorted by rules addresses.
 * I suppose that this speedups linear searches.
 */
struct kip6fw_rules_list kip6fw_rules_active;

uint64_t	kip6fw_bcnt_max; /* Max value of IPv6 Firewall byte counter. */

int		debug_ip6fw;	/* ip6fw:debug_ip6fw */

static struct ip6_fw *fw_buf = NULL; /* Buffer for read kernel IP6FW table. */

static int	sd = -1;	/* Descriptor for socket. */
static size_t	nbytesalloc;

/*
 * Init IP6FW support.
 */
int
kip6fw_init(void)
{
	struct ip6_fw tmp;

#if __FreeBSD_version < 420001
	logmsgx(IPA_LOG_WARNING, "Don't use more than 1024 IPv6 Firewall rules on your system.");
#endif

	if ( (sd = socket(AF_INET6, SOCK_RAW, IPPROTO_RAW)) < 0) {
		logmsg(IPA_LOG_ERR, "kip6fw_init: socket(AF_INET6, SOCK_RAW, IPPROTO_RAW)");
		return -1;
	}

	nbytesalloc = IP6FW_INCREASE_ENTRIES * sizeof tmp;
	if ( (fw_buf = memfunc->mem_malloc(nbytesalloc, m_buf)) == NULL) {
		logmsgx(IPA_LOG_ERR, "kip6fw_init: mem_malloc failed");
		return -1;
	}

	/* Assume, that IP6FW_BCNT is unsigned. */
	switch (sizeof(IP6FW_BCNT(tmp))) {
	case 4:
		kip6fw_bcnt_max = UINT32_MAX;
		break;
	case 8:
		kip6fw_bcnt_max = UINT64_MAX;
		break;
	default:
		logmsgx(IPA_LOG_ERR, "kip6fw_init: unsupported size of fw_bcnt (%lu bytes) in struct ip6_fw",
		    (u_long)sizeof(tmp.fw_bcnt));
		return -1;
	}

	if (debug_ip6fw) {
		logmsgx(IPA_LOG_INFO, "kip6fw_init: raw socket opened, FD %d", sd);
		logmsgx(IPA_LOG_INFO, "kip6fw_init: sizeof(struct ip6_fw) = %lu, sizeof(fw_bcnt) = %lu",
		    (u_long)sizeof(tmp), (u_long)sizeof(tmp.fw_bcnt));
	}

	return 0;
}

/*
 * Dump IP6FW rules to log.
 */
static void
kip6fw_dump_table(void)
{
#define TXT_BUF_LEN 60
	char	txt_buf[TXT_BUF_LEN + 1];
	int	len, total_len = 0;
	u_int	subnumber = 0;
	int	prevnum = -1;
	char	one_rule[1 + 5 + 1 + 10 + 1];
	struct ip6_fw *fw;
	/*
	 *  1 - ' '
	 *  5 - length of rule number (u_short)
	 *  1 - '.'
	 * 10 - length of subnumber (u_int)
	 *  1 - '\0'
	 */

	logmsgx(IPA_LOG_INFO, "kip6fw_dump_table: dump IP6FW table:");
	for (fw = fw_buf;; IP6FWP_NEXT_RULE(fw)) {
		if (prevnum == IP6FWP_NUMBER(fw))
			++subnumber;
		else {
			prevnum = IP6FWP_NUMBER(fw);
			subnumber = 0;
		}
		if (subnumber == 0)
			len = snprintf(one_rule, sizeof one_rule, " %hu", IP6FWP_NUMBER(fw));
		else
			len = snprintf(one_rule, sizeof one_rule, " %hu.%u", IP6FWP_NUMBER(fw), subnumber);
		if (len < 0)
			logmsg(IPA_LOG_ERR, "kip6fw_dump_table: snprintf failed");
		else if (len + 1 > (int)sizeof(one_rule))
			logmsgx(IPA_LOG_WARNING, "kip6fw_dump_table: not enough space (%d chars) in one_rule", len + 1);
		else {
			len = strlen(one_rule);
			if (total_len + len > TXT_BUF_LEN) {
				logmsgx(IPA_LOG_INFO, "  **%s", txt_buf);
				total_len = 0;
			}
			strcpy(txt_buf + total_len, one_rule);
			total_len += len;
		}
		if (IP6FWP_NUMBER(fw) == IP6FW_NUMBER_MAX)
			break;
	}
	if (total_len != 0)
		/* Log the rest of txt_buf. */
		logmsgx(IPA_LOG_INFO, "  **%s", txt_buf);
#undef TXT_BUF_LEN
}

/*
 * Read whole IP6FW rules table from the kernel and find needed
 * rules in this table.
 */
int
kip6fw_get_stat(void)
{
	u_short	number;
	u_int	subnumber;
	socklen_t nbytes;
	struct ip6_fw *fw;
	struct kip6fw_rule *kip6fw_rule;

	nbytes = nbytesalloc;
#if __FreeBSD_version >= 420001
	if (getsockopt(sd, IPPROTO_IPV6, IPV6_FW_GET, fw_buf, &nbytes) < 0
	    && errno != EINVAL) {
		logmsg(IPA_LOG_ERR, "kip6fw_get_stat: getsockopt(IPV6_FW_GET)");
		logmsgx(IPA_LOG_ERR, "kip6fw_get_stat: IPv6 Firewall is not configured in the kernel or does not work properly");
		return -1;
	}
	if (debug_ip6fw)
		logmsgx(IPA_LOG_INFO, "kip6fw_get_stat: nbytes = %lu, nbytesalloc = %lu",
		    (u_long)nbytes, (u_long)nbytesalloc);
	if (nbytes >= nbytesalloc) {
		/* There are some data more in the IP6FW kernel table. */
		while (nbytes >= nbytesalloc) {
			nbytes = nbytesalloc += IP6FW_INCREASE_ENTRIES * sizeof *fw_buf;
			if ( (fw_buf = mem_realloc(fw_buf, nbytesalloc, m_buf)) == NULL) {
				logmsgx(IPA_LOG_ERR, "kip6fw_get_stat: mem_realloc failed");
				return -1;
			}
			if (getsockopt(sd, IPPROTO_IPV6, IPV6_FW_GET, fw_buf, &nbytes) < 0
			    && errno != EINVAL) {
				logmsg(IPA_LOG_ERR, "kip6fw_get_stat: getsockopt(IPV6_FW_GET)");
				logmsgx(IPA_LOG_ERR, "kip6fw_get_stat: IPv6 Firewall is not configured in the kernel or does not work properly");
				return -1;
			}
			if (debug_ip6fw)
				logmsgx(IPA_LOG_INFO, "kip6fw_get_stat: increase table: nbytes = %lu, nbytesalloc = %lu",
				    (u_long)nbytes, (u_long)nbytesalloc);
		}
	} else if (nbytesalloc - nbytes > IP6FW_FREE_ENTRIES * sizeof *fw_buf) {
		/* Reallocate memory if we have too much. */
		nbytesalloc = nbytes + (IP6FW_FREE_ENTRIES - 1) * sizeof *fw_buf;
		if ( (fw_buf = mem_realloc(fw_buf, nbytesalloc, m_buf)) == NULL) {
			logmsgx(IPA_LOG_ERR, "kip6fw_get_stat: mem_realloc failed");
			return -1;
		}
		if (debug_ip6fw)
			logmsgx(IPA_LOG_INFO, "kip6fw_get_stat: decrease table: nbytesalloc = %lu",
			    (u_long)nbytesalloc);
	}
#else
	/* Buggy code, but let's follow sbin/ip6fw.c:list() code. */
	if (getsockopt(sd, IPPROTO_IPV6, IPV6_FW_GET, fw_buf, &nbytes) < 0) {
		logmsg(IPA_LOG_ERR, "kip6fw_get_stat: getsockopt(IPV6_FW_GET)");
		logmsgx(IPA_LOG_ERR, "kip6fw_get_stat: IPv6 Firewall is not configured in the kernel or does not work properly");
		return -1;
	}
#endif /* __FreeBSD_version >= 420001 */

	if (debug_ip6fw)
		kip6fw_dump_table();

	/*
	 * fw_buf is sorted by numbers by the kernel,
	 * kip6fw_rules_active is kept sorted by numbers and subnumbers.
	 * We need only two loops to find needed IP6FW rules.
	 */
	fw = fw_buf;
	for (kip6fw_rule = TAILQ_FIRST(&kip6fw_rules_active); kip6fw_rule != NULL;) {
		if (IP6FWP_NUMBER(fw) > kip6fw_rule->number) {
			/*
			 * No such number in fw_buf, all subnumbers will be
			 * caught by this if() as well.
			 */
			kip6fw_rule->ip6_fw = NULL;
			kip6fw_rule = TAILQ_NEXT(kip6fw_rule, link);
			continue;
		}
		if (IP6FWP_NUMBER(fw) < kip6fw_rule->number) {
			/*
			 * Our number is greater than current number in fw_buf,
			 * skip some rules in fw_buf until equal or greater one
			 * is found.
			 */
			for (IP6FWP_NEXT_RULE(fw); IP6FWP_NUMBER(fw) != IP6FW_NUMBER_MAX; IP6FWP_NEXT_RULE(fw))
				if (IP6FWP_NUMBER(fw) >= kip6fw_rule->number)
					break;

			if (IP6FWP_NUMBER(fw) > kip6fw_rule->number) {
				/*
				 * Greater number was found, all subnumbers
				 * will be caught by the first if().
				 */
				kip6fw_rule->ip6_fw = NULL;
				kip6fw_rule = TAILQ_NEXT(kip6fw_rule, link);
				continue;
			}
		}

		/* At this point we found matched number in fw_buf. */

		/* Save current number in sense of the same subnumbers. */
		number = IP6FWP_NUMBER(fw);
		if (kip6fw_rule->subnumber == 0) {
			/*
			 * Common case for rule number.0
			 * (this if() is for optimization).
			 */
			kip6fw_rule->ip6_fw = fw;
			kip6fw_rule = TAILQ_NEXT(kip6fw_rule, link);
			if (kip6fw_rule != NULL)
				if (kip6fw_rule->number != number)
					/* Next rule isn't number.y. */
					continue;
		}

		/*
		 * Get next rule in fw_buf, this is correct for last IP6FW rule,
		 * because IP6FW_NUMBER_MAX rule cannot have subnumbers, so we
		 * will not be in this point for the last rule in fw_buf.
		 */
		IP6FWP_NEXT_RULE(fw);

		/* Start from the number.1, because number.0 was checked above. */
		for (subnumber = 1; kip6fw_rule != NULL;) {
			if (kip6fw_rule->number != number)
				/*
				 * Out of subnumbers.  Return to the
				 * first if().
				 */
				break;
			if (IP6FWP_NUMBER(fw) == number) {
				/* There is rule with the same number. */

				if (kip6fw_rule->subnumber == subnumber) {
					/* Subnumber is found. */
					kip6fw_rule->ip6_fw = fw;
					kip6fw_rule = TAILQ_NEXT(kip6fw_rule, link);
				} else {
					/*
					 * We don't check if kip6fw_rule->subnumber
					 * is greater than current subnumber, because
					 * everything is sorted and it is greater
					 * than current subnumber.
					 */
				}

				/* Get next kernel rule. */
				IP6FWP_NEXT_RULE(fw);

				/*
				 * Get next subnumber, it will be used only
				 * if next kernel rule has the same number.
				 */
				++subnumber;
			} else { /* IP6FWP_NUMBER(fw) > number */
				/*
				 * Can't find subnumber.  Let's handle
				 * the rest of subnumbers in the first if().
				 */
				kip6fw_rule->ip6_fw = NULL;
				kip6fw_rule = TAILQ_NEXT(kip6fw_rule, link);
				break;
			}
		}
		/* At this point kip6fw_rule is NULL or has next number. */
	}
	return 0;
}

/*
 * Deinit IP6FW support.
 */
int
kip6fw_deinit(void)
{
	mem_free(fw_buf, m_buf);

	if (sd >= 0)
		if (close(sd) < 0) {
			logmsg(IPA_LOG_ERR, "kip6fw_deinit: close(FD %d)", sd);
			return -1;
		}

	return 0;
}
